@@ -1,4 +1,4 @@
-#!/usr/bin/python
+# !/usr/bin/python
 #
 # Copyright (C) 2012 Yoav Aviram.
 #
@@ -6,7 +6,7 @@
 # you may not use this file except in compliance with the License.
 # You may obtain a copy of the License at
 #
-#      http://www.apache.org/licenses/LICENSE-2.0
+# http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS,
@@ -18,6 +18,7 @@ from itertools import islice
 
 import bottlenose
 from lxml import objectify, etree
+import dateutil.parser
 
 
 # https://kdp.amazon.com/help?topicId=A1CT8LK6UW2FXJ
@@ -27,6 +28,7 @@ DOMAINS = {
     'DE': 'de',
     'ES': 'es',
     'FR': 'fr',
+    'IN': 'in',
     'IT': 'it',
     'JP': 'co.jp',
     'UK': 'co.uk',
@@ -42,6 +44,18 @@ class AmazonException(Exception):
     pass
 
 
+class CartException(AmazonException):
+    """Cart related Exception
+    """
+    pass
+
+
+class CartInfoMismatchException(CartException):
+    """HMAC, CartId and AssociateTag did not match
+    """
+    pass
+
+
 class AsinNotFound(AmazonException):
     """ASIN Not Found Exception.
     """
@@ -79,24 +93,67 @@ class BrowseNodeLookupException(AmazonException):
 
 
 class AmazonAPI(object):
-    def __init__(self, aws_key, aws_secret, aws_associate_tag, region="US"):
+    def __init__(self, aws_key, aws_secret, aws_associate_tag, **kwargs):
         """Initialize an Amazon API Proxy.
 
+        kwargs values are passed directly to Bottlenose. Check the Bottlenose
+        API for valid values (some are provided below).
+        For legacy support, the older 'region' value is still supported.
+        Code should be updated to use the Bottlenose 'Region' value
+        instead.
+
         :param aws_key:
             A string representing an AWS authentication key.
         :param aws_secret:
             A string representing an AWS authentication secret.
         :param aws_associate_tag:
             A string representing an AWS associate tag.
-        :param region:
-            A string representing the region, defaulting to "US" (amazon.com)
+
+        Important Bottlenose arguments:
+        :param Region:
+            ccTLD you want to search for products on (e.g. 'UK'
+            for amazon.co.uk).
             See keys of bottlenose.api.SERVICE_DOMAINS for options, which were
             CA, CN, DE, ES, FR, IT, JP, UK, US at the time of writing.
+            Must be uppercase. Default is 'US' (amazon.com).
+        :param MaxQPS:
+            Optional maximum queries per second. If we've made an API call
+            on this object more recently that 1/MaxQPS, we'll wait
+            before making the call. Useful for making batches of queries.
+            You generally want to set this a little lower than the
+            max (so 0.9, not 1.0).
+            Amazon limits the number of calls per hour, so for long running
+            tasks this should be set to 0.9 to ensure you don't hit the
+            maximum.
+            Defaults to None (unlimited).
+        :param Timeout:
+            Optional timeout for queries.
+            Defaults to None.
+        :param CacheReader:
+            Called before attempting to make an API call.
+            A function that takes a single argument, the URL that
+            would be passed to the API, minus auth information,
+            and returns a cached version of the (unparsed) response,
+            or None.
+            Defaults to None.
+        :param CacheWriter:
+            Called after a successful API call. A function that
+            takes two arguments, the same URL passed to
+            CacheReader, and the (unparsed) API response.
+            Defaults to None.
         """
+        # support older style calls
+        if 'region' in kwargs:
+            kwargs['Region'] = kwargs['region']
+            del kwargs['region']
+
+        if 'Version' not in kwargs:
+            kwargs['Version'] = '2013-08-01'
+
         self.api = bottlenose.Amazon(
-            aws_key, aws_secret, aws_associate_tag, Region=region)
+            aws_key, aws_secret, aws_associate_tag, **kwargs)
         self.aws_associate_tag = aws_associate_tag
-        self.region = region
+        self.region = kwargs.get('Region', 'US')
 
     def lookup(self, ResponseGroup="Large", **kwargs):
         """Lookup an Amazon Product.
@@ -105,7 +162,7 @@ class AmazonAPI(object):
             An instance of :class:`~.AmazonProduct` if one item was returned,
             or a list of  :class:`~.AmazonProduct` instances if multiple
             items where returned.
-        """ 
+        """
         response = self.api.ItemLookup(ResponseGroup=ResponseGroup, **kwargs)
         root = objectify.fromstring(response)
         if root.Items.Request.IsValid == 'False':
@@ -201,12 +258,240 @@ class AmazonAPI(object):
         items = AmazonSearch(self.api, self.aws_associate_tag, **kwargs)
         return list(islice(items, n))
 
+    def cart_create(self, items, **kwargs):
+        """CartCreate.
+        :param items:
+            A dictionary containing the items to be added to the cart.
+            Or a list containing these dictionaries.
+            It is not possible to create an empty cart!
+            example: [{'offer_id': 'rt2ofih3f389nwiuhf8934z87o3f4h',
+            'quantity': 1}]
+
+        :return:
+            An :class:`~.AmazonCart`.
+        """
+
+        if isinstance(items, dict):
+            items = [items]
+
+        if len(items) > 10:
+            raise CartException("You can't add more than 10 items at once")
+
+        offer_id_key_template = 'Item.%s.OfferListingId'
+        quantity_key_template = 'Item.%s.Quantity'
+        i = 0
+
+        for item in items:
+            i += 1
+            kwargs[offer_id_key_template % (i, )] = item['offer_id']
+            kwargs[quantity_key_template % (i, )] = item['quantity']
+
+        response = self.api.CartCreate(**kwargs)
+        root = objectify.fromstring(response)
+
+        return AmazonCart(root)
+
+    def cart_add(self, items, CartId=None, HMAC=None, **kwargs):
+        """CartAdd.
+        :param items:
+            A dictionary containing the items to be added to the cart.
+            Or a list containing these dictionaries.
+            It is not possible to create an empty cart!
+            example: [{'offer_id': 'rt2ofih3f389nwiuhf8934z87o3f4h',
+            'quantity': 1}]
+        :param CartId: Id of Cart
+        :param HMAC: HMAC of Cart, see CartCreate for more info
+        :return:
+            An :class:`~.AmazonCart`.
+        """
+        if not CartId or not HMAC:
+            raise CartException('CartId required for CartClear call')
+
+        if isinstance(items, dict):
+            items = [items]
+
+        if len(items) > 10:
+            raise CartException("You can't add more than 10 items at once")
+
+        offer_id_key_template = 'Item.%s.OfferListingId'
+        quantity_key_template = 'Item.%s.Quantity'
+        i = 0
+
+        for item in items:
+            i += 1
+            kwargs[offer_id_key_template % (i, )] = item['offer_id']
+            kwargs[quantity_key_template % (i, )] = item['quantity']
+
+        response = self.api.CartAdd(CartId=CartId, HMAC=HMAC, **kwargs)
+        root = objectify.fromstring(response)
+
+        new_cart = AmazonCart(root)
+        self._check_for_cart_error(new_cart)
+
+        return new_cart
+
+    def cart_clear(self, CartId=None, HMAC=None, **kwargs):
+        """CartClear. Removes all items from cart
+        :param CartId: Id of cart
+        :param HMAC: HMAC of cart. Do not use url encoded
+        :return: An :class:`~.AmazonCart`.
+        """
+        if not CartId or not HMAC:
+            raise CartException('CartId required for CartClear call')
+        response = self.api.CartClear(CartId=CartId, HMAC=HMAC, **kwargs)
+        root = objectify.fromstring(response)
+
+        new_cart = AmazonCart(root)
+        self._check_for_cart_error(new_cart)
+
+        return new_cart
+
+    def cart_get(self, CartId=None, HMAC=None, **kwargs):
+        """CartGet fetches existing cart
+        :param CartId: see CartCreate
+        :param HMAC: see CartCreate
+        :return: An :class:`~.AmazonCart`.
+        """
+        if not CartId or not HMAC:
+            raise CartException('CartId required for CartGet call')
+        response = self.api.CartGet(CartId=CartId, HMAC=HMAC, **kwargs)
+        root = objectify.fromstring(response)
+
+        cart = AmazonCart(root)
+        self._check_for_cart_error(cart)
+
+        return cart
+
+    def cart_modify(self, items, CartId=None, HMAC=None, **kwargs):
+        """CartAdd.
+        :param items:
+            A dictionary containing the items to be added to the cart.
+            Or a list containing these dictionaries.
+            example: [{'cart_item_id': 'rt2ofih3f389nwiuhf8934z87o3f4h',
+            'quantity': 1}]
+        :param CartId: Id of Cart
+        :param HMAC: HMAC of Cart, see CartCreate for more info
+        :return:
+            An :class:`~.AmazonCart`.
+        """
+        if not CartId or not HMAC:
+            raise CartException('CartId required for CartModify call')
+
+        if isinstance(items, dict):
+            items = [items]
+
+        if len(items) > 10:
+            raise CartException("You can't add more than 10 items at once")
+
+        cart_item_id_key_template = 'Item.%s.CartItemId'
+        quantity_key_template = 'Item.%s.Quantity'
+        i = 0
+
+        for item in items:
+            i += 1
+            kwargs[cart_item_id_key_template % (i, )] = item['cart_item_id']
+            kwargs[quantity_key_template % (i, )] = item['quantity']
+
+        response = self.api.CartModify(CartId=CartId, HMAC=HMAC, **kwargs)
+        root = objectify.fromstring(response)
+
+        new_cart = AmazonCart(root)
+        self._check_for_cart_error(new_cart)
+
+        return new_cart
+
+    def _check_for_cart_error(self, cart):
+        if cart._safe_get_element('Cart.Request.Errors') is not None:
+            error = cart._safe_get_element(
+                'Cart.Request.Errors.Error.Code').text
+            if error == 'AWS.ECommerceService.CartInfoMismatch':
+                raise CartInfoMismatchException(
+                    'CartGet failed: AWS.ECommerceService.CartInfoMismatch '
+                    'make sure AssociateTag, CartId and HMAC are correct '
+                    '(dont use URLEncodedHMAC!!!)'
+                )
+            raise CartException('CartGet failed: ' + error)
+
+
+class LXMLWrapper(object):
+    def __init__(self, parsed_response):
+        self.parsed_response = parsed_response
+
+    def to_string(self):
+        """Convert Item XML to string.
+
+        :return:
+            A string representation of the Item xml.
+        """
+        return etree.tostring(self.parsed_response, pretty_print=True)
+
+    def _safe_get_element(self, path, root=None):
+        """Safe Get Element.
+
+        Get a child element of root (multiple levels deep) failing silently
+        if any descendant does not exist.
+
+        :param root:
+            Lxml element.
+        :param path:
+            String path (i.e. 'Items.Item.Offers.Offer').
+        :return:
+            Element or None.
+        """
+        elements = path.split('.')
+        parent = root if root is not None else self.parsed_response
+        for element in elements[:-1]:
+            parent = getattr(parent, element, None)
+            if parent is None:
+                return None
+        return getattr(parent, elements[-1], None)
+
+    def _safe_get_element_text(self, path, root=None):
+        """Safe get element text.
+
+        Get element as string or None,
+        :param root:
+            Lxml element.
+        :param path:
+            String path (i.e. 'Items.Item.Offers.Offer').
+        :return:
+            String or None.
+        """
+        element = self._safe_get_element(path, root)
+        if element is not None:
+            return element.text
+        else:
+            return None
+
+    def _safe_get_element_date(self, path, root=None):
+        """Safe get elemnent date.
+
+        Get element as datetime.date or None,
+        :param root:
+            Lxml element.
+        :param path:
+            String path (i.e. 'Items.Item.Offers.Offer').
+        :return:
+            datetime.date or None.
+        """
+        value = self._safe_get_element_text(path=path, root=root)
+        if value is not None:
+            try:
+                value = dateutil.parser.parse(value)
+                if value:
+                    value = value.date()
+            except ValueError:
+                value = None
+
+        return value
+
 
 class AmazonSearch(object):
     """ Amazon Search.
 
     A class providing an iterable over amazon search results.
     """
+
     def __init__(self, api, aws_associate_tag, **kwargs):
         """Initialise
 
@@ -262,7 +547,8 @@ class AmazonSearch(object):
         """
         response = self.api.ItemSearch(ResponseGroup=ResponseGroup, **kwargs)
         root = objectify.fromstring(response)
-        if root.Items.Request.IsValid == 'False':
+        if (hasattr(root.Items.Request, 'Errors') and
+                not hasattr(root.Items, 'Item')):
             code = root.Items.Request.Errors.Error.Code
             msg = root.Items.Request.Errors.Error.Message
             if code == 'AWS.ParameterOutOfRange':
@@ -273,11 +559,7 @@ class AmazonSearch(object):
         return root
 
 
-class AmazonBrowseNode(object):
-
-    def __init__(self, element):
-        self.element = element
-
+class AmazonBrowseNode(LXMLWrapper):
     @property
     def id(self):
         """Browse Node ID.
@@ -287,8 +569,8 @@ class AmazonBrowseNode(object):
         :return:
             ID (integer)
         """
-        if hasattr(self.element, 'BrowseNodeId'):
-            return int(self.element['BrowseNodeId'])
+        if hasattr(self.parsed_response, 'BrowseNodeId'):
+            return int(self.parsed_response['BrowseNodeId'])
         return None
 
     @property
@@ -298,14 +580,14 @@ class AmazonBrowseNode(object):
         :return:
             Name (string)
         """
-        return getattr(self.element, 'Name', None)
+        return getattr(self.parsed_response, 'Name', None)
 
     @property
     def is_category_root(self):
         """Boolean value that specifies if the browse node is at the top of
         the browse node tree.
         """
-        return getattr(self.element, 'IsCategoryRoot', False)
+        return getattr(self.parsed_response, 'IsCategoryRoot', False)
 
     @property
     def ancestor(self):
@@ -314,7 +596,7 @@ class AmazonBrowseNode(object):
         :return:
             The ancestor as an :class:`~.AmazonBrowseNode`, or None.
         """
-        ancestors = getattr(self.element, 'Ancestors', None)
+        ancestors = getattr(self.parsed_response, 'Ancestors', None)
         if hasattr(ancestors, 'BrowseNode'):
             return AmazonBrowseNode(ancestors['BrowseNode'])
         return None
@@ -341,13 +623,13 @@ class AmazonBrowseNode(object):
     A list of this browse node's children in the browse node tree.
     """
         children = []
-        child_nodes = getattr(self.element, 'Children')
+        child_nodes = getattr(self.parsed_response, 'Children')
         for child in getattr(child_nodes, 'BrowseNode', []):
-                children.append(AmazonBrowseNode(child))
+            children.append(AmazonBrowseNode(child))
         return children
 
 
-class AmazonProduct(object):
+class AmazonProduct(LXMLWrapper):
     """A wrapper class for an Amazon product.
     """
 
@@ -357,78 +639,12 @@ class AmazonProduct(object):
         :param item:
             Lxml Item element.
         """
-        self.item = item
+        super(AmazonProduct, self).__init__(item)
         self.aws_associate_tag = aws_associate_tag
         self.api = api
         self.parent = None
         self.region = kwargs.get('region', 'US')
 
-    def to_string(self):
-        """Convert Item XML to string.
-
-        :return:
-            A string representation of the Item xml.
-        """
-        return etree.tostring(self.item, pretty_print=True)
-
-    def _safe_get_element(self, path, root=None):
-        """Safe Get Element.
-
-        Get a child element of root (multiple levels deep) failing silently
-        if any descendant does not exist.
-
-        :param root:
-            Lxml element.
-        :param path:
-            String path (i.e. 'Items.Item.Offers.Offer').
-        :return:
-            Element or None.
-        """
-        elements = path.split('.')
-        parent = root if root is not None else self.item
-        for element in elements[:-1]:
-            parent = getattr(parent, element, None)
-            if parent is None:
-                return None
-        return getattr(parent, elements[-1], None)
-
-    def _safe_get_element_text(self, path, root=None):
-        """Safe get element text.
-
-        Get element as string or None,
-        :param root:
-            Lxml element.
-        :param path:
-            String path (i.e. 'Items.Item.Offers.Offer').
-        :return:
-            String or None.
-        """
-        element = self._safe_get_element(path, root)
-        if element:
-            return element.text
-        else:
-            return None
-
-    def _safe_get_element_date(self, path, root=None):
-        """Safe get elemnent date.
-
-        Get element as datetime.date or None,
-        :param root:
-            Lxml element.
-        :param path:
-            String path (i.e. 'Items.Item.Offers.Offer').
-        :return:
-            datetime.date or None.
-        """
-        value = self._safe_get_element_text(path=path, root=root)
-        if value is not None:
-            try:
-                value = datetime.datetime.strptime(value, '%Y-%m-%d').date()
-            except ValueError:
-                value = None
-
-        return value
-
     @property
     def price_and_currency(self):
         """Get Offer Price and Currency.
@@ -463,7 +679,8 @@ class AmazonProduct(object):
                 currency = self._safe_get_element_text(
                     'OfferSummary.LowestNewPrice.CurrencyCode')
         if price:
-            return float(price) / 100, currency
+            fprice = float(price) / 100 if 'JP' not in self.region else price
+            return fprice, currency
         else:
             return None, None
 
@@ -519,9 +736,34 @@ class AmazonProduct(object):
             Returns of list of authors
         """
         result = []
-        authors = self._safe_get_element('ItemAttributes.Author') or []
-        for author in authors:
-            result.append(author.text)
+        authors = self._safe_get_element('ItemAttributes.Author')
+        if authors is not None:
+            for author in authors:
+                result.append(author.text)
+        return result
+
+    @property
+    def creators(self):
+        """Creators.
+
+        Creators are not the authors. These are usually editors, translators,
+        narrators, etc.
+
+        :return:
+            Returns a list of creators where each is a tuple containing:
+
+                1. The creators name (string).
+                2. The creators role (string).
+
+        """
+        # return tuples of name and role
+        result = []
+        creators = self._safe_get_element('ItemAttributes.Creator')
+        if creators is not None:
+            for creator in creators:
+                role = creator.attrib['Role'] if \
+                    'Role' in creator.attrib else None
+                result.append((creator.text, role))
         return result
 
     @property
@@ -705,6 +947,15 @@ class AmazonProduct(object):
                     'UPCListElement', root=upc_list[0])
         return upc
 
+    @property
+    def color(self):
+        """Color.
+
+        :return:
+            Color (string)
+        """
+        return self._safe_get_element_text('ItemAttributes.Color')
+
     @property
     def sku(self):
         """SKU.
@@ -759,8 +1010,48 @@ class AmazonProduct(object):
         :return:
             Editorial Review (string)
         """
-        return self._safe_get_element_text(
-            'EditorialReviews.EditorialReview.Content')
+        reviews = self.editorial_reviews
+        if reviews:
+            return reviews[0]
+        return ''
+
+    @property
+    def editorial_reviews(self):
+        """Editorial Review.
+
+        Returns a list of all editorial reviews.
+
+        :return:
+            A list containing:
+
+                Editorial Review (string)
+        """
+        result = []
+        reviews_node = self._safe_get_element('EditorialReviews')
+
+        if reviews_node is not None:
+            for review_node in reviews_node.iterchildren():
+                content_node = getattr(review_node, 'Content')
+                if content_node is not None:
+                    result.append(content_node.text)
+        return result
+
+    @property
+    def languages(self):
+        """Languages.
+
+        Returns a set of languages in lower-case.
+        :return:
+            Returns a set of languages in lower-case (strings).
+        """
+        result = set()
+        languages = self._safe_get_element('ItemAttributes.Languages')
+        if languages is not None:
+            for language in languages.iterchildren():
+                text = self._safe_get_element_text('Name', language)
+                if text:
+                    result.add(text.lower())
+        return result
 
     @property
     def features(self):
@@ -771,9 +1062,10 @@ class AmazonProduct(object):
             Returns a list of 'ItemAttributes.Feature' elements (strings).
         """
         result = []
-        features = self._safe_get_element('ItemAttributes.Feature') or []
-        for feature in features:
-            result.append(feature.text)
+        features = self._safe_get_element('ItemAttributes.Feature')
+        if features is not None:
+            for feature in features:
+                result.append(feature.text)
         return result
 
     @property
@@ -872,3 +1164,107 @@ class AmazonProduct(object):
             return []
 
         return [AmazonBrowseNode(child) for child in root.iterchildren()]
+
+    @property
+    def images(self):
+        """List of images for a response.
+
+        When using lookup with RespnoseGroup 'Images', you'll get a
+        list of images. Parse them so they are returned in an easily
+        used list format.
+        :return:
+            A list of `ObjectifiedElement` images
+        """
+        try:
+            images = [image for image in self._safe_get_element(
+                'ImageSets.ImageSet')]
+        except TypeError:  # No images in this ResponseGroup
+            images = []
+        return images
+
+
+class AmazonCart(LXMLWrapper):
+    """Wrapper around Amazon shopping cart.
+       Allows iterating over Items in the cart.
+    """
+
+    @property
+    def cart_id(self):
+        return self._safe_get_element_text('Cart.CartId')
+
+    @property
+    def purchase_url(self):
+        return self._safe_get_element_text('Cart.PurchaseURL')
+
+    @property
+    def amount(self):
+        return self._safe_get_element_text('Cart.SubTotal.Amount')
+
+    @property
+    def formatted_price(self):
+        return self._safe_get_element_text('Cart.SubTotal.FormattedPrice')
+
+    @property
+    def currency_code(self):
+        return self._safe_get_element_text('Cart.SubTotal.CurrencyCode')
+
+    @property
+    def hmac(self):
+        return self._safe_get_element_text('Cart.HMAC')
+
+    @property
+    def url_encoded_hmac(self):
+        return self._safe_get_element_text('Cart.URLEncodedHMAC')
+
+    def __len__(self):
+        return len(self._safe_get_element('Cart.CartItems.CartItem'))
+
+    def __iter__(self):
+        items = self._safe_get_element('Cart.CartItems.CartItem')
+        if items is not None:
+            for item in items:
+                yield AmazonCartItem(item)
+
+    def __getitem__(self, cart_item_id):
+        """
+        :param cart_item_id: access item by CartItemId
+        :return: AmazonCartItem
+        """
+        for item in self:
+            if item.cart_item_id == cart_item_id:
+                return item
+        raise KeyError('no item found with CartItemId: %s' % (cart_item_id,))
+
+
+class AmazonCartItem(LXMLWrapper):
+    @property
+    def asin(self):
+        return self._safe_get_element_text('ASIN')
+
+    @property
+    def quantity(self):
+        return self._safe_get_element_text('Quantity')
+
+    @property
+    def cart_item_id(self):
+        return self._safe_get_element_text('CartItemId')
+
+    @property
+    def title(self):
+        return self._safe_get_element_text('Title')
+
+    @property
+    def product_group(self):
+        return self._safe_get_element_text('ProductGroup')
+
+    @property
+    def formatted_price(self):
+        return self._safe_get_element_text('Price.FormattedPrice')
+
+    @property
+    def amount(self):
+        return self._safe_get_element_text('Price.Amount')
+
+    @property
+    def currency_code(self):
+        return self._safe_get_element_text('Price.CurrencyCode')
